/***************************************************************************
 *   Copyright (C) 2006 by Jens Koehler                                    *
 *   irimi_at_gmx_de                                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <cstdio>
#include <string>
#include <iostream>
#include <fstream>
#include <cctype>

#include "commands.h"
#include "amarokserver.h"
#include "debugTracer.h"



namespace // static components
{

// can be used for future releases !
char *TAG_IGNORE_LIST []=
    {
        "<Directory>",
        "<Score>",
        "<Type>",
        "<Playcount>",
        "<LastPlayed>",
        "<Filesize>",
        "<Composer>",
        "<DiscNumber>",
        "<Mood>",
        "<SampleRate>",
        "<?xml version"
    };

const char* ITEM_URL_TAG ="<item ";
const char* ITEM_TAG ="<item>";
const char* ITEM_END_TAG= "</item>";
const char* PLAYLIST_END_TAG="</playlist>";
//const char* LIST_END_TAG="</list>";
const char* SUB_TAG ="<sub>";
const char* END_TAG ="</";
const char* XML_AND_CODE="&amp;";

const char* DCOP_BASIC_CMD ="dcop amarok ";
const int MAX_LOG_RETRIES = 5;
};

using namespace std;

AmarokServer::AmarokServer(const char *user,const char *pw ) : p_user (user), p_pw (pw), b_login(false)
{}

AmarokServer::~AmarokServer()
{}

bool AmarokServer::checkVersion()
{
    vector<string> tmp;

    if (popen("dcop amarok player version", tmp) && tmp.size()==1 && tmp[0].size()>=2)
    {
        const char *p=tmp[0].c_str();

        if (p[2]>='4')
            return true;

        cerr << "wrong amarok version:" << tmp[0] << endl;
    }

    tracer.errorOut("ERROR: Couldn't attach to amaroK 1.4.x");
    return false;
}


/** just a very simple solution */
bool AmarokServer::handleLogin (const std::string &userpasswd)
{
    string check (p_user);
    check.append(",");
    check.append(p_pw);
    b_login = (check == userpasswd);
#ifdef VDRAMGW_DEBUG

    if (!b_login)
    {
        tracer.debugOut (check.c_str());
        tracer.debugOut (userpasswd.c_str());
    }
#endif

    return b_login;
}

bool AmarokServer::dcopCommand(const string &cmd_parameter, vector<string> &result)
{
    string cmd = DCOP_BASIC_CMD;
    cmd.append(cmd_parameter);
    return popen(cmd,result);
}

bool AmarokServer::popen ( const string &command, vector<string> &result)
{
#ifdef VDRAMGW_DEBUG
    tracer.debugOut(command.c_str());
#endif
    FILE*           fp;
    const int       BUFFERSIZE = 500;
    char            buf [BUFFERSIZE];
    string      strLine;

    bool bRet=true;
    result = vector<string> ();
    if ((fp = ::popen(command.c_str (), "r")) == NULL)
    {
        tracer.errorOut("popen cannot be created");
        return false;
    }

    while (fgets(buf, sizeof (buf), fp))
    {
        strLine = buf;
#ifdef VDRAMGW_DEBUG

        tracer.debugOut(strLine.c_str());
#endif

        //    if (strLine [strLine.size () - 1] != '\n')
        //    {
        //      tracer.errorOut("BUFFERSIZE too small:" + BUFFERSIZE );
        //      bRet = false;
        //    }
        //    else
        //    {
        if (strLine.find("no such function")==0)
        {
            tracer.errorOut("Dcop:no such function");
            bRet = false;
        }
        else
        {   // remove \n
            result.push_back (strLine.substr (0, strLine.length()- 1));
        }
        //    }
    }
    if (pclose(fp) == -1)
    {
        tracer.errorOut("Cannot execute pclose");
        bRet = false;
    }
    return bRet;
}

bool AmarokServer::readFile (vector<string> &result)
{

    string      strLine;
    string      strFile  = result[0];

    bool bRet=true;
    result = vector<string> ();

    ifstream in(strFile.c_str());
    if (!in)
    {
        tracer.errorOut("Cannot open xml playlist");
        bRet = false;
    }
    else
    {
        while(in && getline(in,strLine) && !in.eof())
        {
            if (!checkIgnoreReduceData(strLine))
                result.push_back (strLine);
        }
    }
    return bRet;

}

void AmarokServer::addMedias(const vector<string> &list)
{
    for(unsigned int ii=0 ; ii<list.size(); ii++)
    {
        //tracer.debugOut (list[ii].c_str());
        addMedia(list[ii]);
    }
}

bool AmarokServer::addMedia(const string &url)
{
    vector<string> result;
    string cmd=DCOP_BASIC_CMD;
    cmd.append("playlist addMedia \"");
    cmd.append(url);
    cmd.append("\"");
    return popen(cmd, result);

}

bool AmarokServer::queryAmarok(const string &query,vector<string> &result)
{

    return popen("dcop amarok collection query \"" +query+"\"", result);

}


void AmarokServer::addArtists(const vector<string> &artistList)
{

    for(unsigned int i=0 ; i<artistList.size(); i++)
    {

        if (artistList[i][0]!=' ')
        {
            vector<string>urlList;
            string query = "SELECT DISTINCT t.url FROM album al, artist ar, tags t WHERE t.artist = ar.id AND t.album = al.id AND ar.name = '";
            query.append(artistList[i]);
            query.append("' order by al.name , t.track");
            queryAmarok(query,urlList);
            addMedias(urlList);
        }
        else
            continue;
    }

}


void AmarokServer::addAlbums(const vector<string> &albumList)
{

    for(unsigned int i=0 ; i<albumList.size(); i++)
    {

        if (albumList[i][0]!=' ')
        {
            vector<string>urlList;
            string query = "SELECT DISTINCT t.url FROM album al, artist ar, tags t WHERE t.artist = ar.id AND t.album = al.id  AND  al.name = '";
            query.append(albumList[i]);
            query.append("' order by t.track");
            queryAmarok(query,urlList);
            addMedias(urlList);
        }
        else
            continue;
    }

}

void AmarokServer::addTitles(const std::vector<std::string> &titleList)
{
    for(unsigned int i=0 ; i<titleList.size(); i++)
    {

        if (titleList[i][0]!=' ')
        {
            vector<string>urlList;
            // @todo new query definition for titles
            string query = "SELECT DISTINCT t.url FROM album al, artist ar, tags t WHERE t.artist = ar.id AND t.album = al.id  AND  al.name = '";
            query.append(titleList[i]);
            query.append("' order by t.track");
            queryAmarok(query,urlList);
            addMedias(urlList);
        }
        else
            continue;
    }

}

void AmarokServer::addGenres (const std::vector<std::string> &genreList)
{
    for(unsigned int i=0 ; i<genreList.size(); i++)
    {

        if (genreList[i][0]!=' ')
        {
            vector<string>urlList;
            // @todo new query definition for genres
            
            string query ="SELECT DISTINCT t.url FROM genre gr, artist ar, tags t WHERE t.artist = ar.id AND t.genre = gr.id  AND  gr.name = '";
            query.append(genreList[i]);
            query.append("' order by t.artist");
            queryAmarok(query,urlList);
            addMedias(urlList);
        }
        else
            continue;
    }

}

/*
bool AmarokServer::queryCurrentTrackStatistics(vector<string> &result)
{
 
  vector<string> tmp;
 
  if (popen("dcop amarok player encodedURL", tmp) && tmp.size()==1)
  {
    string query = "SELECT playcounter, percentage, rating FROM statistics WHERE url = '";
    query.append(tmp[0]);
    query.append("'");
    return popen("dcop amarok collection query \""+query+"\"", result);
  }
 
  return false;
 
}
 
*/
bool AmarokServer::queryArtist(const string &qArtist, vector<string> &result)
{
    string query = "SELECT DISTINCT artist.name FROM tags INNER JOIN artist ON artist.id=tags.artist WHERE tags.sampler = 0 ";
    if (qArtist.size())
    {
        query.append("and artist.name like '%");
        query.append(qArtist);
        query.append("%'");
    }
    query.append(" ORDER BY LOWER( artist.name ) ");
    return popen("dcop amarok collection query \""+query+"\"", result);
}

bool AmarokServer::queryAlbumArtist(const string &qAlbum, vector<string> &result)
{
    string query = "SELECT DISTINCT al.name from album al, artist ar, tags t where t.artist = ar.id and t.album = al.id and ar.name = '";
    query.append(qAlbum);
    query.append("' ORDER BY LOWER( al.name ) ");

    return popen("dcop amarok collection query \""+query+"\"", result);
}

bool AmarokServer::queryGenreArtist(const string &qGenre, vector<string> &result)
{
    string query = "SELECT DISTINCT ar.name from genre gr, artist ar, tags t where t.artist = ar.id and t.genre=gr.id and gr.name='";
    query.append(qGenre);
    query.append("' ORDER BY LOWER( ar.name ) ");

    return popen("dcop amarok collection query \""+query+"\"", result);
}

bool AmarokServer::queryGenre(const string &qGenre, vector<string> &result)
{
    string query = "SELECT DISTINCT genre.name FROM tags INNER JOIN genre ON genre.id=tags.genre WHERE tags.sampler = 0 ";
    if (qGenre.size())
    {
        query.append("and genre.name like '%");
        query.append(qGenre);
        query.append("%'");
    }
    query.append(" ORDER BY LOWER( genre.name ) ");

    return popen("dcop amarok collection query \""+query+"\"", result);
}



bool AmarokServer::queryAlbum(const string &qAlbum, vector<string> &result)
{
    string query ="SELECT DISTINCT album.name FROM album ORDER  BY lower( album.name )";
    if (qAlbum.size())
    {
        query.append("and album.name like '%");
        query.append(qAlbum);
        query.append("%'");
    }
    query.append(" ORDER BY LOWER( al.name ) ");
    return popen("dcop amarok collection query \""+query+"\"", result);
}



/**filter data we are not interested in to reduce needed bandwith
and remove xml end tag */
bool AmarokServer::checkIgnoreReduceData(string &str)
{
    // remove leading spaces
    unsigned int index =str.find_first_not_of (" ");
    str=str.substr(index,str.length()-index);
    // do nothing
    if (0 == str.find(ITEM_END_TAG) || 0 == str.find(PLAYLIST_END_TAG) )
        return false;

    // current str in ignore list to drop?
    for(unsigned int i = 0; i < sizeof(TAG_IGNORE_LIST) / sizeof(char*); i++)
    {
        if ( 0 == str.find(TAG_IGNORE_LIST[i]) )
            return true;
    }
    //@todo redefine xml syntax "&amp;" -> "&"
    // redefine item tag
    if (str.find(ITEM_URL_TAG)==0)
        str.assign(ITEM_TAG);
    else // remove xml end tag
    {
        str= str.substr (0, str.find(END_TAG));
    }
    return false;
}


void AmarokServer::runServer(int port)
{


    bool bReconnect = true;
    vector<string> cmdList;
    try
    {
        TCPSocketWrapper sockserver;

        vector<string> result;
        sockserver.listen(port);

        tracer.infoOut("server is ready");

        // accept connection from client
        while (bReconnect)
        {
            tracer.infoOut("server is waiting for new connection ...");

            int retries=0;
            TCPSocketWrapper sock(sockserver.accept());

            tracer.infoOut("accepted connection from:");
            tracer.infoOut(sock.address());

            TCPStream stream(sock);

            bool doitagain = true;
            int int_cmd = CMD_ERROR;
            string cmd_parameter;

            while (doitagain && retries < MAX_LOG_RETRIES )
            {
                stream >> int_cmd ; // wait for the command

                if (!b_login && int_cmd != CMD_REQUEST_LOGIN && int_cmd != CMD_CHECK_VERSION )
                {
                    tracer.infoOut("AUTHENTIFICATION REQUIRED");
                    retries++;
                    continue;
                }


                switch (int_cmd)
                {
                case CMD_CHECK_VERSION:
                    do
                    {
                        getline(stream, cmd_parameter);
                    }
                    while (cmd_parameter.empty());
                    if (cmd_parameter.find(VDRAMGW_VERSION))
                    {
                        stream << CMD_ERROR  << endl;
                        string msg="WRONG CLIENT VERSION:";
                        msg.append(cmd_parameter);
                        msg.append("expected is ");
                        msg.append("VDRAMGW_VERSION");
                        tracer.errorOut(msg.c_str());
                        doitagain = false;
                    }
                    else
                        stream << int_cmd  << endl;

                    break;
                case CMD_REQUEST_LOGIN: // login
                    do
                    {
                        getline(stream, cmd_parameter);
                        //cout << cmd_parameter << endl;
                    }
                    while (cmd_parameter.empty());
                    handleLogin (cmd_parameter) ? stream << CMD_REQUEST_LOGIN << endl : stream << CMD_ERROR << endl;
                    break;
                case CMD_REQUEST_ARTISTLIST:
                case CMD_REQUEST_GENRELIST:
                case CMD_REQUEST_ALBUMLIST:
                case CMD_REQUEST_STATISTICS:
                case CMD_REQUEST_LIST:

                    tracer.debugOut("CMD_REQUEST_LIST");
                    do
                    {
                        getline(stream, cmd_parameter);
                    }
                    while (cmd_parameter.empty());
                    listRequest (stream, int_cmd, cmd_parameter, result);
                    break;
                case CMD_REQUEST_ADDARTIST:
                    tracer.debugOut("CMD_REQUEST_ADDARTIST");
                    getParameterRequest(stream,cmdList);
                    stream << int_cmd  << endl; // ALWAYS ACK
                    addArtists(cmdList);
                    cmdList = vector<string> ();
                    break;
                case CMD_REQUEST_ADDGENRE:
                    tracer.debugOut("CMD_REQUEST_ADDGENRE");
                    getParameterRequest(stream,cmdList);
                    stream << int_cmd  << endl; // ALWAYS ACK
                    addGenres(cmdList);
                    cmdList = vector<string> ();
                    break;
                case CMD_REQUEST_ADDALBUM:
                    tracer.debugOut("CMD_REQUEST_ADDALBUM");
                    getParameterRequest(stream,cmdList);
                    stream << int_cmd  << endl; // ALWAYS ACK
                    addAlbums(cmdList);
                    cmdList = vector<string> ();
                    break;
                case CMD_REQUEST_ADDTITLE:
                    tracer.debugOut("CMD_REQUEST_ADDTITLE");
                    getParameterRequest(stream,cmdList);
                    stream << int_cmd  << endl; // ALWAYS ACK
                    addTitles(cmdList);
                    cmdList = vector<string> ();
                    break;
                case CMD_REQUEST_GET:
                    tracer.debugOut("CMD_REQUEST_GET");
                    do
                    {
                        getline(stream, cmd_parameter);
                    }
                    while (cmd_parameter.empty());
                    if (dcopCommand(cmd_parameter,result) && result.size()>=1)
                    {
                        stream << int_cmd  << endl; // ACK
                        stream << result[0] << endl;
                    }
                    else
                    {
                        stream << CMD_ERROR  << endl; // FAILED
                    }
                    break;
                case CMD_REQUEST_PLAYLIST:
                    tracer.debugOut("CMD_REQUEST_PLAYLIST");
                    do
                    {
                        getline(stream, cmd_parameter);
                    }
                    while (cmd_parameter.empty());
                    if (dcopCommand(cmd_parameter,result) && result.size()>=1 )
                    {
                        if (readFile(result))
                        {
                            stream << int_cmd  << endl; // ACK
                            std::vector<std::string>::iterator p=result.begin();
                            while (p != result.end())
                            {
                                stream << *p << endl;
                                p++;
                            }
                            break;
                        }
                    }
                    else
                        stream << CMD_ERROR  << endl;

                    break;
                case CMD_REQUEST_SET:
                    do
                    {
                        getline(stream, cmd_parameter);
                    }
                    while (cmd_parameter.empty());
                    if (dcopCommand(cmd_parameter,result))
                        stream << int_cmd  << endl; // ACK
                    else
                        stream << CMD_ERROR  << endl; // FAILED
                    break;
                case CMD_REQUEST_CLIENTDOWN: // client shutdown
                    tracer.debugOut("CMD_REQUEST_CLIENTDOWN");
                    doitagain = false;
                    break;
                case CMD_REQUEST_SERVERDOWN: // server shutdown
                    tracer.debugOut("CMD_REQUEST_SERVERDOWN");
                    doitagain = false;
                    bReconnect=false;
                    break;

                default:
                    tracer.errorOut("unkown amarok_cmd type");
                    stream << CMD_ERROR  << endl; // FAILED
                    break;

                }
                result = vector<string> ();
                cmd_parameter.clear();
            }
        }
    }
    catch (const SocketRunTimeException &e)
    {
        tracer.errorOut ("socket exception:");
        tracer.errorOut (e.what());

    }

    tracer.infoOut("server is closing socket connection");


}

void AmarokServer::getParameterRequest (TCPStream &stream,vector<string>&parameterList)
{
    string cmd_parameter;
    do
    {
        getline(stream, cmd_parameter);
        if (!cmd_parameter.empty())
        {
            parameterList.push_back(cmd_parameter);
#ifdef VDRAMGW_DEBUG
            tracer.debugOut(cmd_parameter.c_str());
#endif
        }
    }
    while (cmd_parameter.empty());

}
void AmarokServer::listRequest (TCPStream &stream,int int_cmd, const string &cmd_parameter, vector<string>&result)
{
    bool ok= false;

    result.push_back("<list>");

    switch (int_cmd)
    {
    case CMD_REQUEST_STATISTICS:
        //ok=queryCurrentTrackStatistics(result);
        break;
    case CMD_REQUEST_ARTISTLIST:
        ok=buildArtistsResultlist(result);
        break;
    case CMD_REQUEST_ALBUMLIST:
        ok=buildAlbumsResultlist(result);
        break;
    case CMD_REQUEST_GENRELIST:
        ok=buildGenresResultlist(result);
        break;
    case CMD_REQUEST_LIST:
        ok=dcopCommand(cmd_parameter,result);
        break;
    default:
        break;
    }

    result.push_back("</list>");


    if (ok)
    {

        stream << int_cmd  << endl; // ACK
        vector<string>::iterator p=result.begin();
        while (p != result.end())
        {
            stream << *p << endl;
            p++;
        }
    }
    else
        stream << CMD_ERROR  << endl; // FAILED

}

// generate list which is simple to parse
bool AmarokServer::buildArtistsResultlist (vector<string>&result)
{
    bool bRet=true;
    vector<string>artistList;
    vector<string>albumList;

    if(bRet=queryArtist("",artistList)) // get all
    {

        for(unsigned int i=0 ; i<artistList.size(); i++)
        {
            string text=ITEM_TAG;
            text.append(artistList[i]);
            result.push_back(text);
            if(bRet=queryAlbumArtist(artistList[i],albumList))
            {

                for(unsigned int ii=0 ; ii<albumList.size(); ii++)
                {
                    if (albumList[ii].length()>1)
                        result.push_back(SUB_TAG+albumList[ii]);
                }

            }
            else
                break;
        }
    }

    return bRet;
}

// generate list which is simple to parse
bool AmarokServer::buildAlbumsResultlist (vector<string>&result)
{
    bool bRet=true;
    vector<string>albumList;

    if(bRet=queryAlbum("",albumList)) // get all
    {

        for(unsigned int i=0 ; i<albumList.size(); i++)
        {
            string text=ITEM_TAG;
            text.append(albumList[i]);
            //cout << text << endl;
            result.push_back(text);
        }
    }

    return bRet;
}

// generate list which is simple to parse
bool AmarokServer::buildGenresResultlist (vector<string>&result)
{
    bool bRet=true;
    vector<string>genreList;
    vector<string>artistList;

    if(bRet=queryGenre("",genreList)) // get all
    {

        for(unsigned int i=0 ; i<genreList.size(); i++)
        {
            string text=ITEM_TAG;
            text.append(genreList[i]);
            result.push_back(text);
            if(bRet=queryGenreArtist(genreList[i],artistList))
            {

                for(unsigned int ii=0 ; ii<artistList.size(); ii++)
                {
                    if (artistList[ii].length()>1)
                        result.push_back(SUB_TAG+artistList[ii]);
                }

            }
            else
                break;
        }
    }

    return bRet;
}
