/***************************************************************************
 *   Copyright (C) 2006 by Jens Koehler                                    *
 *   irimi_at_gmx_de                                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _AMAROK_SERVER_
#define _AMAROK_SERVER_

#include <string>
#include <streambuf>
#include <vector>

#include "../Sockets.h"
#include "debugTracer.h"



class AmarokServer
{
public:

    AmarokServer( const char* user, const char* pw);
    virtual ~AmarokServer();
    void runServer(int port);
    bool checkVersion();

    DebugTracer tracer;

private:
    bool addMedia(const std::string &url);
    void addMedias(const std::vector<std::string> &list);

    void addArtists(const std::vector<std::string> &artistList);
    void addAlbums (const std::vector<std::string> &albumList);
    void addTitles(const std::vector<std::string> &titleList);
    void addGenres (const std::vector<std::string> &genreList);

    bool queryAmarok(const std::string &query,std::vector<std::string> &result);
    bool queryArtist(const std::string &qArtist,std::vector<std::string> &result);
    bool queryAlbumArtist(const std::string &qAlbum,std::vector<std::string> &result);
    bool queryGenreArtist(const std::string &qGenre, std::vector<std::string> &result);
    bool queryAlbum(const std::string &qGenre,std::vector<std::string> &result);
    bool queryGenre(const std::string &qGenre,std::vector<std::string> &result);


    bool checkIgnoreReduceData(std::string &strLine);
    bool dcopCommand(const std::string &, std::vector<std::string> &result);
    bool popen ( const std::string &command, std::vector<std::string> &result);
    bool readFile (std::vector<std::string> &result);
    bool handleLogin(const std::string &userpasswd);

    void listRequest (TCPStream &stream, int cmd, const std::string &parameter, std::vector<std::string>&result);
    bool buildArtistsResultlist (std::vector<std::string> &result);
    bool buildAlbumsResultlist (std::vector<std::string> &result);
    bool buildGenresResultlist (std::vector<std::string> &result);
    void getParameterRequest (TCPStream &stream,std::vector<std::string>&parameterList);

    const char *p_user;
    const char *p_pw;
    bool b_login;

};

#endif /*_AMAROK_SERVER_*/
